#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stream.h>
#include "Ygl.h"
#include "syntax.h"
#include "box.h"
#include "vector.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"
#include "fplot.h"
#include "pde.cc"
#include "global.cc"

//**************************************************************************************************

void header() {
     fprintf(stderr,"\n******************************************************");
     fprintf(stderr,"\n*                                                    *");
     fprintf(stderr,"\n*  Calcium diffusion/buffering solver: version 4.3.1 *");
     fprintf(stderr,"\n*    Victor Matveev, MRB/NIDDK/NIH, March 1, 2002    *");
     fprintf(stderr,"\n*                                                    *");
     fprintf(stderr,"\n******************************************************\n");
}

//**************************************************************************************************

#define  EXTRA_PARAM_STRING  "; pA=5.182134 ; pi=4 atan(1) ; "

extern int VERBOSE;
extern int DIMENSIONALITY;
extern int GEOMETRY;
extern const int SPHERICAL;

void printResults(class TokenString &Param, 
                  class FieldObj *Ca, class BufferArray *Bufs, class KineticObj *KO);

//**************************************************************************************************

class LoopObj
{
public:

  char loopVarString[ 2048 ];

  PlotArray *plots;
  VectorObj *result;

  int       num, steps;
  VectorObj var, var0, dvar;
  char      **ids, **trackIDs, *tp;

  LoopObj(TokenString &, VectorObj *);
  ~LoopObj();
 
  int  operator()() { return num; }
  void step(int);
  void draw();
};

int getTrackVarNum(TokenString &);
int getTrackVar(TokenString &TS, int n, int * = 0, int * = 0);

//**************************************************************************************************
//                                           M A I N
//**************************************************************************************************


int main(int argc, char **argv) {

  if (argc < 2) { fprintf(stderr, "*** Usage: %s parameter_file \n",argv[0]); exit(1); }
                     
  TokenString *TS = new TokenString(argv[1], EXTRA_PARAM_STRING, argc, argv);    // Parse the parameter file   
  TS->get_int_param("verbose", 0, &VERBOSE);
  if (VERBOSE) header();

  if (TS->token2_count("geometry","spherical") || TS->token3_count("geometry",ASSIGN_LABEL,"spherical")) 
      { DIMENSIONALITY = 1; GEOMETRY = SPHERICAL; } 
  else if  (TS->token2_count("geometry","conical") || TS->token3_count("geometry",ASSIGN_LABEL,"conical"))
      { DIMENSIONALITY = 2; GEOMETRY = 2; } 

  VectorObj result( getTrackVarNum(*TS) );
  LoopObj   vary( *TS, &result );
  delete TS;
  char temp[512];
  int  pos;

  // ++++++++++++++++++++++++++++ MAIN LOOP ++++++++++++++++++++++++++++

  for (int step = 0; step <= vary.steps; step++) {
    
    vary.step(step);
    char *extra = new char [ strlen(vary.loopVarString) + strlen( EXTRA_PARAM_STRING ) + 1];
    strcpy(extra, vary.loopVarString); strcat( extra, EXTRA_PARAM_STRING);
    TokenString Params(argv[1], extra, argc, argv);  // Parse the parameter file at each iteration  
    delete [] extra;
    if (Params.token_count("exit")) break;
    Params.get_int_param("verbose", 0, &VERBOSE);

    if ( !Params.token_count("Run") ) {    //################ CALCULATOR MODE ##############
      if (VERBOSE) fprintf(stderr, "***** No simulation runs specified; running in calculator mode *****\n");
      if ( vary() ) for (int i = 0; i < result.size; i++) 
         result[i] = Params.get_double( getTrackVar(Params, i) );
      printResults(Params, 0, 0, 0);
    }        // beware of undestructed string
    else if ( ! equal(Params.get_string("mode"), "ODE") )  //#######  FULL PDE MODE  #######
      {
      RegionObj   synapse(Params);         // Read pre-synaptic geometry
      FieldObj::set_region(&synapse);      // Bind fields to geometry variable "synapse"
      set_bc_types(Params);                // Read boundary condition definitions
      set_dimensions(Params);              // read grid dimensions from parameter file
      grid_stretch(Params);                // make non-uniform grid (read parameters from file)
      grid_clamp(Params);                  // clamp grid nodes to specified locations
      initScheme();                        // set up finite difference support arrays
      FieldObj Ca(Params);                 // set up calcium field (read parameters from file)
      BufferArray Bufs(Params);            // set up buffer fields (read parameters from file)
      initTortuosity(Params, Ca, Bufs);    // initialize space-dependent diffusion coefficient:
      KineticObj Gates(Params, &Ca, &Bufs);// set up Ca binding kinetic variables

      if ( Params.token_count("Import", &pos) ) Import( Params.line_string( pos + 1, temp ), &Gates, &Ca, &Bufs);
      PlotArray *plots = new PlotArray(Params, &Ca, &Bufs, &Gates);

      if (!Ca.source_num) fprintf(stderr,"\n>>> Error: no Ca%c channels specified <<<\n",178), exit(1);
      Run(Params, Ca, Bufs, Gates, *plots);//##### RUN THE DIFFERENCE SCHEME #####
      if ( vary() ) for (int i = 0; i < result.size; i++) 
        result[i] = *ResolveID( vary.trackIDs[i], 0, &Ca, &Bufs, &Gates, "Main: Cannot track undefined");
      printResults(Params, &Ca, &Bufs, &Gates);
      delete plots;
      //for (int ti = 0; ti < DiffNum; ti ++) fprintf(stderr,"\n TA[%d]=%ld \n",ti, DiffArray[ti]);
      killTortuosity();
      kill_grid();
      kill_bc();
      }
    else   //#####################  ODE solver mode: #################
      {
      if (VERBOSE) fprintf(stderr, "\n***** Integrating the ODEs only *****\n");     
      KineticObj Gates(Params, 0, 0);            // set up ODEs
      if ( Params.token_count("Import", &pos) ) Import( Params.line_string( pos + 1, temp), &Gates, 0, 0);
      PlotArray *plots = new PlotArray(Params, 0, 0, &Gates);
      RunODEs(Params, Gates, plots);
      if ( vary() ) for (int i = 0; i < result.size; i++)  
        result[i] = *ResolveID( vary.trackIDs[i], 0, 0, 0, &Gates, "Main: Cannot track undefined");
      printResults(Params, 0, 0, &Gates);
      delete plots;
      }

    vary.draw(); 
  }                                   

}

//**********************************************************************************************
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//**********************************************************************************************


int getTrackVarNum(TokenString &TS) {
  int num = 0;
  for (int i=0; i<TS.token_count("track"); i++) {
    num += TS.tokens_to_eol( TS.token_index("track", i+1) + 1);
  }
  return num;
}

int getTrackVar(TokenString &TS, int n, int *iGraph, int *iSets) {  // n = 0..track var number - 1
  int num = 0, m;
  for (int i=0; i<TS.token_count("track"); i++) {
    if (iGraph) *iGraph = i;
    int p = TS.token_index("track", i+1) + 1; 
    num += ( m = TS.tokens_to_eol( p ) );
    if (iSets) *iSets =  m;
    if (num >= n + 1) return p + m - ( num - n );
  }
  fprintf(stderr, "*** ERROR in getTrackVar(): something is terribly wrong\n");
  exit(1);
}


//**********************************************************************************************

LoopObj::LoopObj(TokenString &TS, VectorObj *res) : num( TS.token_count(LOOP_TOKEN) ), 
         var(num), var0(num), dvar(num), ids(new char *[num]), tp(new char[num])
  {
  int i;
  if (!num) { steps = 0; return; }

  // if (!res->size)
  //     fprintf(stderr,"## LoopObj::LoopObj: no variables to track in the \"for\" loop\n"), exit(1);

  steps = 100000;
  char prefix[256], filename[256];

  strcpy(prefix, "");
  result = res;

  double v0, v1, dv = 1.0;

  trackIDs = new char *[res->size + 1];

  for (i = 0; i < res->size; i++) trackIDs[i] = TS.StrCpy( getTrackVar(TS, i) );

  for (i = 0; i < num; i++)
      {       
      ids[i] = TS.StrCpy( TS.token_index(LOOP_TOKEN, i+1) + 1 ); 
      v0 = KinExpression(TS, TS.token3_index(LOOP_TOKEN, ids[i],ASSIGN_LABEL) + 1, "Error in \"for\" loop").Evaluate();
      v1 = KinExpression(TS, TS.token_index("to", i+1) + 1, "Error in \"for\" loop").Evaluate();
      dv = KinExpression(TS, TS.token_index("step", i+1) + 1, "Error in \"for\" loop").Evaluate();

      if ( ( fabs(v0 - int(v0+0.5)) < 1.0e-6 ) && 
           ( fabs(v1 - int(v1+0.5)) < 1.0e-6 ) &&  
           ( fabs(dv - int(dv+0.5)) < 1.0e-6 ) )  
      tp[i] = 'i'; else tp[i] = 'd';

      var.elem[i] = var0.elem[i] = v0;
      dvar.elem[i] = dv;
   
      int n = int( (v1 - v0) / dv + 0.5 );
      if (n < steps) steps = n;
    
      fprintf(stderr, "\n### For loop: for(%s = %g; %s <= %g; %s += %g)\n",
	      ids[i], v0, ids[i], v1, ids[i], dv);
      }

  char *xlabel = new char[strlen(ids[0]) + 3];
  strcpy(xlabel, ids[0]);

  if ( TS.token2_count("plot.method","xmgr") ) {
     plots = new PlotArray(result->size + TS.token_count("track"));
     plots->method = METHOD_XMGR;
     XmgrPlot::init(TS, TS.token_count("track"));
  }
  else if (TS.token2_count("plot.method","Ygl")) {
     plots = new PlotArray(2 * result->size);
     plots->method = METHOD_YGL; 
     plots->ygl_on = 1;
     strcat(xlabel,"=");
  }
  else { 
     plots = new PlotArray(result->size);
     plots->method = METHOD_MUTE;
  }
 
  int pos;
  if (TS.token_count("plot.print", &pos)) TS.line_string(pos + 1, prefix);

  int xmgrGraph, xmgrGraph0 = -1, graphNum, xmgrSets;

  for (i = 0; i < result->size; i++) {
    fprintf(stderr, "    tracking variable #%d: %s\n", i+1, trackIDs[i]);
    sprintf(filename,"%s%s", prefix, trackIDs[i]);

    if (plots->method == METHOD_YGL) {
      plots->set_plot( new PointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps, trackIDs[i], xlabel, 
                                       pref_width, pref_height, trackIDs[i]),"point");
      plots->set_plot( new MutePointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps, filename, trackIDs[i]), "point file");
      }
    else if ( plots->method == METHOD_MUTE )
      plots->set_plot( new MutePointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps,  filename, trackIDs[i]), "point file");
    else if ( plots->method == METHOD_XMGR ) {
      getTrackVar(TS, i, &xmgrGraph, &xmgrSets);
      // fprintf(stderr,"setnum=%d\n",xmgrSets);
      if (xmgrGraph > xmgrGraph0) {
         plots->set_plot( new XmgrPointPlot(xmgrSets, 0, steps * dv, xlabel), "point xmgr");
         graphNum = ( xmgrGraph0 =  xmgrGraph ) + i;
      }
      ((XmgrPointPlot *)(plots->array[graphNum]))->set_set( (*result)()+i, var()+num-1, trackIDs[i]);
      plots->set_plot( new MutePointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps,  filename, trackIDs[i]), "point file");
    }
  }

  }

//**************************************************************************************************
//**************************************************************************************************

  LoopObj::~LoopObj()
  {
   int i;
   if (!num) return;

   delete plots;
   delete [] tp;

   for (i=0; i<result->size; i++) delete trackIDs[i];
   for (i=0; i<num; i++)          delete ids[i];

   delete [] ids; delete [] trackIDs;
 
  }


//**************************************************************************************************

void LoopObj::step(int iStep) {
  
  int i;
  char temp[1024];
  strcpy( loopVarString, "");

  if (!num) return;

  var = var0 + iStep * dvar;

  for (i = 0; i < num; i++)
    {       
    if (tp[i] == 'i')
      sprintf(temp, "%s = %d", ids[i], int(var[i] + 0.5) );
    else
      sprintf(temp, "%s = %g", ids[i], var[i] );

    strcat( loopVarString, temp );
    if ( i < num - 1 ) strcat( loopVarString, "; " );
    }
 
  if (VERBOSE) {
    fprintf(stderr,  "\n    Next step in the \"for\" loop: %s \n", loopVarString);
    fprintf(stderr,  "========================================================\n");
  }
 
  //strcat( loopVarString, "; " );
}

//**************************************************************************************************

  void LoopObj::draw()
    {
    int i;
    if (!num) return;

    plots->draw_all();  

    if (VERBOSE && result->size) {
      fprintf(stderr,"\n========================================================\n ");
      fprintf(stderr,"    %s : ", loopVarString);
      for (i = 0; i < result->size; i++) fprintf(stderr," %s=%g", trackIDs[i], (*result)[i]);
      fprintf(stderr,"\n========================================================\n");
    }
    fflush(stdout);
    }

//**************************************************************************************************

void printResults(class TokenString &Params, 
                  class FieldObj *Ca, class BufferArray *Bufs, class KineticObj *KO) {

  int    i, p;
  char   *fileStr;
  FILE   *f;
  char   temp[1024];

  for (i = 0; i < Params.token_count("print"); i++) {
    p = Params.token_index("print", i+1) + 2;
    fileStr = Params.get_string(p - 1);
    //fprintf(stderr,"s=%s p0=%d num=%d count=%d\n",s,p0,num,Params.token_count("print"));

    if ( ::equal(fileStr,"stderr") ) f = stderr;
    else if ( ::equal(fileStr,"stdout") ) f = stdout;
    else f = fopen(fileStr,"w");
    if (!f) { fprintf(stderr, ">>> printResults: couldn't open file %s\n",fileStr); return; }

    fprintf( f,"%s\n", Params.line_string( p, temp, Ca, Bufs, KO) );
    if ( (f != stdout) && (f != stderr) ) fclose(f); 
  }


  for (i = 0; i < Params.token_count("append"); i++) {

    p = Params.token_index("append", i+1) + 2;
    fileStr = Params.get_string(p - 1);

    if ( ::equal(fileStr,"stderr") ) f = stderr;
    else if ( ::equal(fileStr,"stdout") ) f = stdout;
    else f = fopen(fileStr,"a");
    if (!f)  { fprintf(stderr, ">>> printResults: couldn't open file %s\n",fileStr); return; }

    fprintf( f,"%s\n", Params.line_string( p, temp, Ca, Bufs, KO) );
    if ( (f != stdout) && (f != stderr) ) fclose(f); 
  }
}

//**************************************************************************************************

