#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"

extern double CHARGE_LOSS;
extern int  VERBOSE;
extern long location_to_index(double, double, double);


//**************************************************************************

inline double binary(char op, double &a, double b)
{
  switch(op) {
   case T_PLUS:  return a += b;
   case T_MINUS: return a -= b;
   case T_OR:    return a = ( (a > 0) || (b > 0) ? 1: 0 );
   case T_AND:   return a = ( (a > 0) && (b > 0) ? 1: 0 );
   case T_GT:    return a = (a > b ? 1 : 0);
   case T_GE:    return a = (a >= b ? 1 : 0);
   case T_LT:    return a = (a < b ? 1 : 0);
   case T_LE:    return a = (a <= b ? 1 : 0);
   case T_EQ:    return a = (a == b ? 1 : 0);
   case T_NEQ:   return a = (a != b ? 1 : 0);
   case T_MULT:  return a *= b;
   case T_DIV:   return b == 0.0 && b == 0.0 ? 1.0 : a /= b;
   case T_MOD:   return a = double ( int(a) % int(b) );
   case T_POWER: return a = pow(a,b);
    
   default: fprintf(stderr,"!!! Operation [%d=%c] not recognized\n", op, op);
            exit(1);
  }
}

//**************************************************************************

inline double function(char op, double x)
{
  switch (op)
    {
    case BRACKET:   return x;
    case T_NOT:     return (x > 0) ? 0 : 1;
    case T_INT:     return double(int(x));
    case T_COSH:    return cosh(x);
    case T_SINH:    return sinh(x);
    case T_COS:     return cos(x);
    case T_SIN:     return sin(x);
    case T_TANH:    return tanh(x);
    case T_TAN:     return tan(x);
    case T_ATAN:    return atan(x);
    case T_EXP:     return exp(x);
    case T_LOG:     return log(x);
    case T_LOG10:   return log10(x);
    case T_SQR:     return x * x;
    case T_CUBE:    return x * x * x;
    case T_SQRT:    return sqrt(x);
    case T_THETA:   if (x == 0.0 ) return 0.5; else return (x > 0) ? 1 : 0;
    case T_SIGMA:   return 0.5 * (1 + tanh(x) );
    default:        fprintf(stderr,"\n>>> Error in expression: function \"%c\" not recognized\n", op);
                    exit(1);
    }
}

//*************************************************************************
/*
inline void pushOp(char op, double *valStack, int &valHead, char *opStack, int &opHead ) {

  // fprintf(stderr, ">> pushing operator #%d = %s (valStack=%g valHead=%d opStack=%d opHead=%d)\n",
  //                op,type_id[op],valStack[valHead],valHead,opStack[opHead],opHead);

  char pr = priority[ op ];

  while ( opHead && priority[ opStack[opHead - 1] ] <= pr ) {
    opHead --; valHead --;
    binary( opStack[opHead], valStack[opHead], valStack[opHead+1] );
  }

  opStack[ opHead++ ] = op;
}
 */
//**************************************************************************

double ExpressionObj::Evaluate(int *firstTerm)
  {
  struct TermStruct term;
  char   pr, unaryOp = 0;
  char   opStack[priorityLevels];
  double val, valStack[priorityLevels + 1];
  int    j, j0, stackHead = 0;

  if (term_num <= 0 ) return 0.0;
  if (firstTerm) j0 = *firstTerm;
  else j0 = 0;

  for (j = j0; j < term_num; j ++) { // *** LOOP OVER TERMS BEGINS HERE ***

    char Type = (term = term_array[j]).type;

    if ( isUnary(Type) ) {   // Catch unary sign operators here
      unaryOp = term.type; continue;
    }

    if ( isBinary(Type) ) {  // ...So this binary operator is not a unary operator
      pr = priority[ Type ];
      while ( stackHead )
        if ( priority[ opStack[stackHead - 1] ] <= pr ) {
          stackHead --; 
          binary( opStack[stackHead], valStack[stackHead], valStack[stackHead+1] );
	} else break;
      opStack[ stackHead++ ] = Type;
      //pushOp( Type, valStack, valHead, opStack, stackHead ); 
      continue;
    }

    if ( Type == BR_CLOSE )  break;   // catch a closing bracket
                                                       // *** The term is a number:
    if ( isFunction(Type) ) {                          // 1. a function call (or nested expression)
      j ++;
      val = function( Type, this->Evaluate(&j) );
    }
    else if ( Type == NUMBER_TYPE )  val = term.val;   // 2. a float      
    else if ( Type == POINTER_TYPE ) val = *term.ptr;  // 3. a variable
    else { fprintf(stderr,"\n>>> ERROR: Cannot evaluate this expression: unknown term %d\n", Type); 
           print(stderr); exit(1); }

    if (unaryOp) {
      if  (unaryOp == T_UNARY_NOT)  val = (val > 0) ? 0 : 1;
      else if (unaryOp == T_UNARY_MINUS)  val = -val;
      unaryOp = 0;
    }

    valStack[ stackHead ] = val;
  }

  if (firstTerm) *firstTerm = j;
  while ( stackHead-- ) 
    binary( opStack[stackHead], valStack[stackHead], valStack[stackHead+1] ); // *** Collect the result

  return valStack[0] ;
  }
   	            
//**************************************************************************

int ExpressionObj::eliminate(int kill0, int kill1) {

  //  fprintf(stderr,">>> eliminating terms #%d (%d) through #%d (%d); max term number = %d\n", 
  //        kill0, kill1, term_array[kill0].type, term_array[kill1].type, term_num) ;

  if (kill0 < 0 || kill1 >= term_num) {
    fprintf(stderr,">>> Cannot kill terms #%d through #%d (max term number = %d)\n", kill0, kill1, term_num) ;
    exit(1);
  }

  int d = kill1 - kill0 + 1;
  for (int i = kill1 + 1; i < term_num; i++) term_array[i-d] = term_array[i];
  term_num -= d;
  //fprintf(stderr,"!"); fflush(stderr);
  return d;
}
 
// *************************************************************************

void ExpressionObj::pushOp(char op, char *opStack, int *indStack, int &stackHead, int &j) {

  //fprintf(stderr, ">> pushing operator #%d = %s (indStack=%d opStack=%d stackHead=%d)\n",
  //                op,type_id[op],indStack[stackHead],opStack[stackHead],stackHead);

  char pr = ( op == BR_CLOSE ) ? priorityLevels : priority[ op ];
  char reduce;

  while ( stackHead )
    if ( priority[ reduce = opStack[stackHead - 1] ] <= pr ) {
      if ( stackHead > 1 )
	if ( priority[ opStack[stackHead - 2] ] < priority[ reduce ] ) break; // don't eliminate [stackHead - 1] 
      // binary operation if it had lower priority than the one before that - this means there was a pointer
      // involved, otherwise [stackHead - 2] would have been eliminated itself

      //fprintf(stderr,"stackHead=%d opStack[stackHead - 1]=%d, priority=%d, op=%d, pr=%d\n",
      //        stackHead, reduce, priority[ reduce ], op, pr);

      int j0 = j;
      int j1 = indStack[stackHead-1];          // Points to the _end_ of the 1st operand
      int j2 = indStack[stackHead];           // Points to the _end_ of the 2nd operand
      struct TermStruct t1 = term_array[j1];
      struct TermStruct t2 = term_array[j2];

      // fprintf(stderr,"t1=%d, j1=%d, val1=%g   t2=%d j2=%d val2=%g\n",t1.type,j1,t1.val,t2.type,j2,t2.val);

      if ( t1.type == NUMBER_TYPE && t2.type == NUMBER_TYPE ) {  // *** Binary with two constants
        binary( reduce, t1.val, t2.val );
        term_array[j1] = t1;
        j -= eliminate(j1+1, j2);
      }
      else if ( t1.type == NUMBER_TYPE ) {          // *** First operand is a number
        if (t1.val == 0.0 ) {                                // *** Check 0.0
          if ( reduce == T_MULT || reduce == T_DIV  || reduce == T_AND)
            j -= eliminate(j1+1, j2);                        // 0 (*,/,and) ptr = 0
          else if (reduce == T_PLUS || reduce == T_OR)       // 0 (+,or) ptr = ptr
            j -= eliminate(j1, j1+1);
          else if (reduce == T_MINUS)                        // 0 - ptr = - ptr
            j -= eliminate(j1, j1);
	}
        else if (t1.val == 1.0) {                            // *** Check 1.0
          if (reduce == T_MULT || reduce == T_AND) {         // 1 (*,and) ptr = ptr
            j -= eliminate(j1, j1);
            if ( term_array[j1].type == T_MULT || term_array[j1].type == T_AND ) j -= eliminate(j1, j1);
	  }
          else if (reduce == T_OR)                           // 1 or ptr = 1
            j -= eliminate(j1+1, j2);
	}
      }
      else if ( t2.type == NUMBER_TYPE ) {          // *** Second operand is a number
        if (t2.val == 0.0 ) {                                // *** Check 0
          if ( reduce == T_MULT || reduce == T_AND )  {      // ptr (*,and) 0 = 0
            if ( term_array[j2-1].type == T_MULT || term_array[j2-1].type == T_AND ) j -= eliminate(j2-1, j2-1);
            if ( t1.type != BR_CLOSE ) j -= eliminate(j1, j1);
            else while ( !isFunction( term_array[j1].type ) ) j -= eliminate(j1, j1);
	  }
          else if ( reduce == T_PLUS || reduce == T_MINUS || reduce == T_OR )
            j -= eliminate(j1+1, j2);                        // ptr (+,-,or) 0 = ptr
          else if ( reduce == T_POWER ) {                    // ptr^0 = 1
            term_array[j2].val = 1.0;
            j -= eliminate(j2-1, j2-1);
            if ( t1.type != BR_CLOSE ) j -= eliminate(j1, j1);
            else while ( !isFunction( term_array[j1].type ) ) j -= eliminate(j1, j1);
	  }
          else if ( reduce == T_DIV || reduce == T_MOD  ) {  // ptr /% 0 = error
            fprintf(stderr,">>> Error in optimizer: division by zero in expression");
            print(stderr); exit(1);
	  }
	}
        else if (t2.val == 1.0) {                            // *** Check 1.0
          if (reduce == T_MULT || reduce == T_DIV || reduce == T_MOD || reduce == T_AND || reduce == T_POWER ) {
            // fprintf(stderr,"eliminating exponentiation to 1\n");
            // fprintf(stderr,"t1=%d, j1=%d, val1=%g   t2=%d j2=%d val2=%g\n",t1.type,j1,t1.val,t2.type,j2,t2.val);
            j -= eliminate(j1+1, j2);  }                      // ptr (*,/,mod,and,^) 1 = ptr
          else if (reduce == T_OR) {                         // ptr or 1 = 1
            j -= eliminate(j2-1, j2-1);
            if ( t1.type != BR_CLOSE ) j -= eliminate(j1, j1);
            else while ( !isFunction( term_array[j1].type ) ) j -= eliminate(j1, j1);
	  }
	}
      }
      if ( j < j0 ) stackHead --;
      else break;
    }  else break;

  opStack[ stackHead++ ] = op;
}

//**************************************************************************

void ExpressionObj::Optimize(int *firstTerm)
  {
  struct TermStruct term;
  char   Type, unaryOp = 0;
  char   *opStack = new char[term_num];
  int    *indStack = new int[term_num+1];
  int    j, j0, stackHead = 0;

  if (term_num <= 0 ) { delete [] opStack; delete [] indStack; return; }
  if (firstTerm) j0 = *firstTerm;
  else j0 = 0;

  for (j = j0; j < term_num; j ++) { // *** LOOP OVER TERMS BEGINS HERE ***

    Type = (term = term_array[j]).type;
    
    //fprintf(stderr, "Type=%d\n",Type);

    if ( isUnary(Type) ) {   // Catch unary sign operators here
      unaryOp = term.type; continue;
    }

    if ( isBinary(Type) ) {  // ...So this binary operator is not a unary operator
      pushOp( Type, opStack, indStack, stackHead, j); 
      continue;
    }

    if ( Type == BR_CLOSE )  break;    // catch a closing bracket
                                                       // *** The term is a number:
    if ( isFunction(Type) ) {         // 1. a function call (or nested expression)
      int jold = j; j ++;
      Optimize(&j); 
      if (jold + 2 == j  && ( term_array[jold+1].type == NUMBER_TYPE || Type == BRACKET ) ) { // kill a function call
        //fprintf(stderr,"killing f call\n");
        j -= eliminate(jold, jold);
        if (Type != BRACKET) term_array[jold].val = function( Type, term_array[jold].val );
        j -= eliminate(jold+1, jold+1);
        Type = term_array[jold].type;
      }
    }
    else if ( Type != NUMBER_TYPE && Type != POINTER_TYPE)  {
       fprintf(stderr,"\n>>> ERROR in optimizer: Cannot evaluate this expression: unknown term %d\n", Type); 
       print(stderr); exit(1); 
    }

    if ( unaryOp != 0 && Type == NUMBER_TYPE) {     // kill a unary operator
      if  (unaryOp == T_UNARY_NOT)  term_array[j].val = (term_array[j].val > 0) ? 0 : 1;
      else if (unaryOp == T_UNARY_MINUS)  term_array[j].val = -term_array[j].val;
      j -= eliminate(j-1, j-1);
    }
    unaryOp = 0;

    indStack[ stackHead ] = j;
    //fprintf(stderr," ind=%d term=(%d,%g)\n",j,term_array[j].type,term_array[j].val);
  }

  pushOp( BR_CLOSE, opStack, indStack, stackHead, j); // *** Collect the result
  if (firstTerm) *firstTerm = j;

  delete [] opStack; delete [] indStack;
 }
   	            
//**************************************************************************
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//***************************************************************************

bool isFloat(const char *s)
{
 bool pointFlag = false, expFlag = false;
 int  last = strlen(s) - 1;
 int  i;

 for (i=0; i < strlen(s); i++) {
   if ( is_numeral(s[i]) ) continue;
   else if ( s[i] == '.' ) { 
     if ( pointFlag || expFlag || (last == 0) ) return false; // a single dot character is not a number
     else pointFlag = true; 
   } 
   else if ( is_sign(s[i]) ) {
     if ( i == last ) return false;
     if ( (i == 0) && !( is_numeral(s[1]) || (s[1] == '.') ) ) return false;  // mantissa sign
     if ( (i > 0) && ( !( s[i-1] == 'e' || s[i-1] == 'E' ) || !is_numeral(s[i+1]) ) ) return false; // exponent sign
   }
   else if ( s[i] == 'e' || s[i] == 'E' ) {
     if ( (i == last) || (i == 0) ) return false;
     if ( !is_numeral(s[i-1]) || !(is_numeral(s[i+1]) || is_sign(s[i+1])) ) return false;
     if (expFlag) return false;
     expFlag = true;
   }
   else return false;
 }
 return true;
}

//***************************************************************************

bool isInt(const char *s)
{
 int  i;

 for (i=0; i < strlen(s); i++) {
   if ( is_numeral(s[i]) ) continue;
   else if ( is_sign(s[i]) ) { if (i > 0 ) return false; }
   else return false;
 }
 return true;
}

//**************************************************************************

KinExpression::KinExpression(TokenString &Param, int p0, const char *message,
                             double *arg1, const char *arg1id, 
                             double *arg2, const char *arg2id, 
                             double *arg3, const char *arg3id,  
                             FieldObj *Ca, BufferArray *Bufs, KineticObj *KO, int *pLast) :
  ExpressionObj( 2 * get_max_term_num(Param, p0) )
{
  //VERBOSE = 9;

  int tp;
  int p = p0;
  int term = 0;
  int bracket_num = 0;

  if ( term_num == 0 && pLast == 0 ) {
     fprintf(stderr, "\n>>> %s:\n>>> Syntax error - empty expression (token #%d):\n", message, p0);
     Param.linePrint(p0);
     exit(1);      
  }

  while ( !isExpressionStop( Param[p] ) )  { // *************  Main expression loop  ***************

    if (VERBOSE > 6)  fprintf(stderr,"expression: term(%d) = %s \n", term , Param[p] );

    for (tp = 0; tp < TYPE_NUM; tp++)
      if (Param.equal(p, type_id[tp])) {  // it's an operator/function token
	if (VERBOSE > 6)  fprintf(stderr,"it's an operator token = %s\n",type_id[tp]);
        term_array[term].type = tp; 

        if ( Param.equal(p, "+") || Param.equal(p, "-") )  { // convert a binary "+"/"-" to unary, if appropriate
          if ( term == 0 )  term_array[term].type = (tp += (T_UNARY_PLUS - T_PLUS));
          else if ( !isOperand(term_array[term-1].type) ) term_array[term].type = (tp += (T_UNARY_PLUS - T_PLUS));
	}

        if ( isBinary(tp) )    // catch a binary operator as first token, or two binaries in a row
          if (term == 0)  term_num = term; 
          else if ( !isOperand(term_array[term-1].type) ) term_num = term;

        if ( isUnary(tp) && term > 0)  // catch a unary operator following an operand
          if ( isOperand(term_array[term-1].type) ) term_num = term;

        if ( tp == BR_CLOSE ) 
          if ( --bracket_num < 0 ) term_num = term;  // bracket mismatch

        if ( tp == BRACKET ||  isFunction( tp ) ) bracket_num ++;

        if ( isFunction( tp ) && term > 0 ) 
          if ( isOperand( term_array[term-1].type ) ) { // add a multiplication
            term_array[term + 1].type = tp;
            term_array[term].type = T_MULT;
            term ++;
	    }
        break;
      }

    if ( tp >= TYPE_NUM) { // did not match any token

      if (Param.isConst( Param[p], &(term_array[term].val)) ) {  // it's a number
        if (VERBOSE > 6)  fprintf(stderr,"it's a number = %g\n", term_array[term].val );
        term_array[term].type = NUMBER_TYPE;
      }
      else if ( term_array[term].ptr = ResolveID( Param[p], 0, Ca, Bufs, KO, 0,
						arg1, arg1id, arg2, arg2id, arg3, arg3id) ) {  // it's a pointer
        if (VERBOSE > 6)  fprintf(stderr,"it's a pointer = %ld\n", &term_array[term].ptr);
        term_array[term].type = POINTER_TYPE;
      }
      else  term_num = term;  // does not match anything

      if ( term > 0 ) if ( isOperand( term_array[term-1].type ) ) { // add a multiplication
            term_array[term + 1] = term_array[term];
            term_array[term].type = T_MULT;
            term ++;
      }
    }

    if (term_num <= term) break;
    p++; term++;
    if ( p >= Param.token_num ) { term_num = term - 1; break; }

  }  // ******************************   Main expression loop 

  if ( isExpressionStop( Param[p] ) )  term_num = term; 

  if ( bracket_num > 0 ) {
       fprintf(stderr, "\n>>> %s:\n>>> Bracket nesting error:\n", message);
       Param.linePrint(p - 1);
       exit(1);      
  }

  if (term_num) {                           // Run some more checks on the last expression:
                                            // expression can only end in a number or a bracket    
    while ( !isOperand( term_array[term_num-1].type ) && (term_num > 1) )  term_num--;

    char last = term_array[term_num-1].type;
    if (term_num == 1)                     // one-term expression is either a pointer or a number
      if ( (last != NUMBER_TYPE ) && (last != POINTER_TYPE ) ) term_num = 0;
  } 

  if (pLast) *pLast = p - 1;
  else if ( !isExpressionStop( Param[p] ) ) {
       fprintf(stderr, "\n>>> %s:\n>>> Cannot parse the expression:\n", message);
       Param.linePrint(p);
       exit(1);      
     }
   
  // ********************************************************************************************** 

  int j = 0;
  Optimize(&j);
  // fprintf(stderr,"optimized \n");

  if (j < term_num) {
       fprintf(stderr, "\n>>> %s:\n>>> Syntax error in expression (probably a bracket nesting error):\n", message);
       Param.linePrint(p0 + j - 1);
       exit(1);      
     }

  buildFormulaString(Ca, Bufs, KO, arg1, arg1id, arg2, arg2id, arg3, arg3id, message);

  if (VERBOSE > 6) fprintf(stderr,"\nExpression = { %s}\n", formula);

  Evaluate();
}

//**************************************************************************

int get_max_term_num(TokenString &Param, int p)
{
  int terms = 0;
  
  while ( !isExpressionStop( Param[p] ) && ( p < Param.token_num ) ) 
     { p++; terms++; }

  return terms;
}

//**************************************************************************


void ExpressionObj::print(FILE *f)
{
  fprintf(f,"%s ", formula);
  fflush(f);
}


//**************************************************************************

void ExpressionObj::buildFormulaString(FieldObj *Ca, BufferArray *Bufs, KineticObj *KO, 
		 double *p1, const char *s1, double *p2, const char *s2, double *p3, const char *s3, const char *message) {
  int  term;
  char tp;

  strcpy(formula,"");

  for (term = 0; term < term_num; term++) {
    tp = term_array[term].type;
    if (tp < TYPE_NUM) { if ( tp != T_MULT ) strcat(formula, type_id[tp]); else continue; }
    else if (tp == NUMBER_TYPE) {
      char value[64];
      sprintf(value,"%g", term_array[term].val );
      strcat(formula, value);
    }
    else if (tp == POINTER_TYPE) {
      char *str = StrCpy( ResolvePtr( term_array[term].ptr, Ca, Bufs, KO, p1, s1, p2, s2, p3, s3) );
      // fprintf(stderr, "formula=%s str=%s\n", formula, str);
      strcat(formula, str);
      delete str;
    }
    else
      fprintf(stderr, "\n>>> An error in expression \"%s\": bad type=%d\n",formula,tp), exit(1); 
    strcat(formula," ");
  }
}
//**************************************************************************

double *ResolveID(char *svar, double **t, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO, const char *message,
		 double *p1, const char *s1, double *p2, const char *s2, double *p3, const char *s3) {
  int j;

  if (VERBOSE > 8) fprintf(stderr, ">> ResolveID: Resolving %s\n", svar);

    if ( equal(svar, s1) ) return p1;

    if ( equal(svar, s2) ) return p2;

    if ( equal(svar, s3) ) return p3;

    if (Ca) {
     if ( equal(svar, Ca->ID) )  
       { if (t) *t = &Ca->Time; return (*Ca)(); }
     if ( equal(svar, "ICa") )  
       { if (t) *t = &Ca->Time; return &(Ca->ICa); }
     if ( equal(svar, "Charge.loss") )  
       { if (t) *t = &Ca->Time; return &(CHARGE_LOSS); }
    }     

    if (Bufs)
      for (j = 0; j < Bufs->buf_num; j++)
	if ( equal(svar,Bufs->array[j]->ID) )  
          { if (t) *t = &Bufs->array[0]->Time;  return (*Bufs->array[j])(); }

    if (KO)
      {
      if (t) *t = &KO->Time;

      if (KO->var) 
        for (j = 0; j < KO->var->get_size(); j++) 
          if (equal(svar, KO->var_id[j])) 
            return (*KO->var)() + j;

      if (KO->ident)
        for (j = 0; j < KO->ident->get_size(); j++)
          if (equal(svar, KO->ident_id[j])) 
            return  (*KO->ident)() + j;

      for (j = 0; j < KO->tables->table_num; j++)
	 if (equal(svar, KO->tables->array[j]->ID ) )
           return &(KO->tables->array[j]->value);

      for (j = 0; j < KO->maxima->peak_num; j++)
	 if (equal(svar, KO->maxima->array[j]->ID ) ) 
           return &(KO->maxima->array[j]->peak);

      for (j = 0; j < KO->locations->interpol_num; j++)
	 if (equal(svar, KO->locations->array[j]->ID ) ) 
	   {
           if (t) *t = &(KO->locations->array[j]->eigentime);
	   return &(KO->locations->array[j]->result);
	   }
      }

    if (message) fprintf(stderr,"\n>>> %s \"%s\"\n", message, svar ), exit(1);
    return 0;
}

//**************************************************************************


const char *ResolvePtr(double *ptr, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO,
		 double *p1, const char *s1, double *p2, const char *s2, double *p3, const char *s3)
{

   int j;
    if (ptr == p1) return s1;
    if (ptr == p2) return s2;
    if (ptr == p3) return s3;


    if (Ca) {
       if ( ptr == Ca->elem )  return Ca->ID; 
       else if ( ptr == &(Ca->ICa) )  return "ICa";
    } 

    if (Bufs)
      for (j = 0; j < Bufs->buf_num; j++)
	if ( ptr == Bufs->array[j]->elem )  return  Bufs->array[j]->ID; 

    if (KO)
      {
      if (KO->var) 
        for (j = 0; j < KO->var->get_size(); j++)
         if ( ptr == ((*KO->var)() + j) ) return KO->var_id[j];
          
      if (KO->ident)
        for (j = 0; j < KO->ident->get_size(); j++)
         if ( ptr == ((*KO->ident)() + j) ) return KO->ident_id[j];

      for (j = 0; j < KO->tables->table_num; j++)
	 if ( ptr == &(KO->tables->array[j]->value) ) return KO->tables->array[j]->ID;

      for (j = 0; j < KO->maxima->peak_num; j++)
	 if ( ptr == &(KO->maxima->array[j]->peak) ) return KO->maxima->array[j]->ID;

      for (j = 0; j < KO->locations->interpol_num; j++)
	 if ( ptr == &(KO->locations->array[j]->result) ) return KO->locations->array[j]->ID;
      }

    return 0;
    //fprintf(stderr,"ResolvePtr: cannot resolve pointer %X\n", ptr );
    //exit(1);
}


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


