#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stream.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "interpol.h"

extern  double *xcoord, *ycoord, *zcoord; 
extern  int    xsize, ysize, zsize;
extern  long   xysize; 
extern  int    VERBOSE;
extern  char   in_bounds(double, double, double);

extern int DIMENSIONALITY;
extern int GEOMETRY;
extern const int SPHERICAL;

//**************************************************************************

double InterpolObj::Evaluate(double t)
   {
     if (average) { 
        eigentime = *tptr; 
        return result = Field->average();
     }
   
     eigentime = t;

     if (*tptr != newtime)   // field has been updated since last interpolation evaluation
       {
       oldtime = newtime;
       oldres = newres;
       newtime = *tptr;
       newres  = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
                *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;
       }

     // fprintf(stderr,"%g %g    %g %g   %g \n",oldtime,oldres,newtime,newres,*tptr);

     if (newtime > oldtime) // temporal extrapolation using values at two different times
       result = ( (newtime - t) * oldres + (t - oldtime) * newres ) / (newtime - oldtime);
     else result = newres;
     
     return result;
   }
//**************************************************************************

void InterpolObj::reset()
   {
   if (average) return;

   result = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
            *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;

   oldres = newres = result;
   oldtime = newtime = *tptr;   
   }

//*******************************************************************************
//*******************************************************************************

InterpolObj::InterpolObj(TokenString &Param, int p0, FieldObj *FO)
   {
   double x=0.0, y=0.0, z=0.0;
   int p = p0, pLast;
   average = false;   
   char xyz[128];
   
   Field = FO;
   double *field = (*FO)();
   tptr = &(FO->Time); 
 
   if ( Param.equal(p + 1, "]") )  {
     strcpy(xyz, Param[p-1]); strcat( xyz, "[]" );
     average = true;
     Evaluate(0.0);
     pLast = p;
     }
   else {

   if ( DIMENSIONALITY > 2 ) {
     x = KinExpression(Param, p + 1, "", 0, "%", 0, "%", 0, "%", 0, 0, 0, &pLast).Evaluate();
     if ( !Param.equal(pLast + 1, ",") ) {
         fprintf(stderr,"*** Syntax error: bad concentration interpolation object\n");
         Param.linePrint(p); exit(1);
     }
     p = pLast + 1;
   }

   if ( DIMENSIONALITY > 1 ) {
     y = KinExpression(Param, p + 1, "", 0, "%", 0, "%", 0, "%", 0, 0, 0, &pLast).Evaluate();
     if ( !Param.equal(pLast + 1, ",") ) {
         fprintf(stderr,"*** Syntax error: bad concentration interpolation object\n");
         Param.linePrint(p); exit(1);
     }
     p = pLast + 1;
   }

   z = KinExpression(Param, p + 1, "", 0, "%", 0, "%", 0, "%", 0, 0, 0, &pLast).Evaluate();
   if ( !Param.equal(pLast + 1, "]") ) {
       fprintf(stderr,"*** Syntax error: bad concentration interpolation object\n");
       Param.linePrint(p); exit(1);
   }
 
   if ( DIMENSIONALITY == 1 ) { x = z; y = z = 0.0; sprintf(xyz,"%s[%g]", FO->ID, x); }
   else if ( DIMENSIONALITY == 2 ) { x = y; y = z; z = 0.0; sprintf(xyz,"%s[%g,%g]", FO->ID, x, y); }
   else sprintf(xyz,"%s[%g,%g,%g]", FO->ID, x, y, z);

   if ( !in_bounds(x, y, z) ) 
     fprintf(stderr,"*** ERROR in InterpolObj::InterpolObj: location [%g,%g,%g] is out of bounds\n", x, y, z);
   }

   ID = StrCpy( xyz );

   token_ptr_source = &(Param.token_ptr[p0-1]);   // restore at destruction time
   token_ptr_target = Param.token_ptr[p0-1];      //  
   Param.token_ptr[p0-1] = ID;                    // *** hijack the token pointer
   Param.deleteToken(p0, pLast + 1);

   if (average) return;

   double p1 = 1.0, p2 = 1.0, p3 = 1.0;
   int    ix = 0, iy = 0, iz = 0;  
   
   while (x >= xcoord[ix] && ix < (xsize - 1) ) ix++;
   if ( DIMENSIONALITY > 1 )  while (y >= ycoord[iy] && iy < (ysize - 1) ) iy++;
   if ( DIMENSIONALITY > 2 )  while (z >= zcoord[iz] && iz < (zsize - 1) ) iz++;

   // fprintf(stderr,"\n(%d,%d,%d)==(%g,%g,%g) point_type=%d\n",
   //        ix,iy,iz,xcoord[ix],ycoord[iy],zcoord[iz], FO->region->point_type(ix,iy,iz) );

   int opt = 0;
   do {    // find an interpolation cube close to (ix,iy,iz) with all vertices 
           // inside the volume
     if ( FO->region->point_type(ix,iy,iz) && FO->region->point_type(ix-1,iy,iz) &&
          FO->region->point_type(ix,iy-1,iz) && FO->region->point_type(ix,iy,iz-1) &&
          FO->region->point_type(ix-1,iy-1,iz) && FO->region->point_type(ix-1,iy,iz-1) &&
          FO->region->point_type(ix,iy-1,iz-1) && FO->region->point_type(ix-1,iy-1,iz-1) )
     break;
     else switch(opt) {

     case 0: ix++; break;      case 1: ix-=2; break;     case 2: iy++; break;     
     case 3: ix++; break;      case 4: ix++; break;      case 5: iy-=2; break;
     case 6: ix--; break;      case 7: ix--; break;      case 8: ix++; iy++; iz++; break;
     case 9: ix++; break;      case 10: ix-=2; break;    case 11: iy++; break;     
     case 12: ix++; break;     case 13: ix++; break;     case 14: iy-=2; break;
     case 15: ix--; break;     case 16: ix--; break;     case 17: ix++; iy++; iz-=2; break;
     case 18: ix++; break;     case 19: ix-=2; break;    case 20: iy++; break;     
     case 21: ix++; break;     case 22: ix++; break;     case 23: iy-=2; break;
     case 24: ix--; break;     case 25: ix--; break;     case 26: ix++; iy++; iz-=2; break;

     case 27: fprintf(stderr,"\n**** InterpolObj failure: check interpolation point!\n"); exit(1);
     }
   } while (++opt);

   /*
   fprintf(stderr," opt=%d (%d,%d,%d)\n", opt, ix, iy, iz);
   fprintf(stderr,"x=%g xcoord[ix-1]=%g xcoord[ix]=%g\n",x, xcoord[ix-1],xcoord[ix]);
   fprintf(stderr,"y=%g ycoord[iy-1]=%g ycoord[iy]=%g\n",y, ycoord[iy-1],xcoord[iy]);
   fprintf(stderr,"z=%g zcoord[iz-1]=%g zcoord[iz]=%g\n",z, zcoord[iz-1],xcoord[iz]);
   */

   p1 = (x - xcoord[ix-1]) / (xcoord[ix] - xcoord[ix-1]);
   if (DIMENSIONALITY > 1) p2 = (y - ycoord[iy-1]) / (ycoord[iy] - ycoord[iy-1]);
   if (DIMENSIONALITY > 2) p3 = (z - zcoord[iz-1]) / (zcoord[iz] - zcoord[iz-1]);

   f000 = f001 = f010 = f011 = f100 = f101 = f110 = f111 = field + ix + iy * xsize + iz * xysize;

   if ( ix > 0 ) { f000 -= 1;      f001 -= 1;      f010 -= 1 ;     f011 -= 1; }
   if ( iy > 0 ) { f000 -= xsize;  f001 -= xsize;  f100 -= xsize ; f101 -= xsize; }
   if ( iz > 0 ) { f000 -= xysize; f010 -= xysize; f100 -= xysize; f110 -= xysize ; }

   p111 =         p1  *      p2  *      p3 ;
   p000 =    (1 - p1) * (1 - p2) * (1 - p3);
   p100 =         p1  * (1 - p2) * (1 - p3);
   p010 =    (1 - p1) *      p2  * (1 - p3);
   p001 =    (1 - p1) * (1 - p2) *      p3 ;
   p110 =         p1  *      p2  * (1 - p3);
   p101 =         p1  * (1 - p2) *      p3 ;
   p011 =    (1 - p1) *      p2  *      p3 ;
 
   result = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
            *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;

   //fprintf(stderr,"p1=%g p2=%g p3=%g\n",p1,p2,p3);
   //fprintf(stderr,"(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)\n",
   //   *f000,p000,*f001,p001,*f010,p010,*f011,p011,*f100,p100,*f101,p101,*f110,p110,*f111,p111);

   oldres = newres = result;
   oldtime = newtime = *tptr;   
}

//*******************************************************************************

InterpolArray::InterpolArray(TokenString &Param, FieldObj *Ca, BufferArray *Bufs)
{
  interpol_num = 0;

  if (Ca) interpol_num += Param.token2_count(Ca->ID, "[");

  if (Bufs)
    for (int j = 0; j < Bufs->buf_num; j++)
        interpol_num += Param.token2_count(Bufs->array[j]->ID, "[");

    array = new InterpolObj *[interpol_num];
    int count = 0;

    if (!interpol_num) return; 

    if (VERBOSE) { fprintf(stderr,"\n### Interpolating: "); fflush(stderr); }

    if (Ca) 
      while ( Param.token2_count(Ca->ID, "[") )
        {
        int p0 = Param.token2_index(Ca->ID, "[", 1);
        array[count] = new InterpolObj(Param, p0, Ca );
        if (VERBOSE) fprintf(stderr," %s",array[count]->ID);
        count ++;
        }
      

    if (Bufs)
      for (int j = 0; j < Bufs->buf_num; j++)
        while ( Param.token2_count(Bufs->array[j]->ID, "[") )
          {
          int p0 = Param.token2_index(Bufs->array[j]->ID, "[", 1);
          array[count] = new InterpolObj(Param, p0, Bufs->array[j] );
          if (VERBOSE) fprintf(stderr," %s",array[count]->ID);
          count ++;
          }

    if (VERBOSE) fprintf(stderr,"\n # Compressing interpolation array: ");

    InterpolObj *temp;

    for (int i = 0; i < interpol_num - 1; i++ )
      for  (int j = i + 1; j < interpol_num; j++ )
	  if ( equal(array[i]->ID, array[j]->ID) )
	    {
	    if (VERBOSE) { fprintf(stderr,"."); fflush(stderr); }
            interpol_num--;
	    temp = array[interpol_num]; 
            array[interpol_num] = array[j];
            array[j] = temp;
	    }

    if (VERBOSE) fprintf(stderr," done\n");
}

//*******************************************************************************

